#ifndef __ROBOT_CONTROLLER__
#define __ROBOT_CONTROLLER__

/* Include Player files */
#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

/* Include GMapping files - Map and OrientedPoint *
 * conflict - so use GMapping:: everywhere        */
#include <utils/commandline.h>
#include <gridfastslam/gridslamprocessor.h>
#include <utils/orientedboundingbox.h>
#include <configfile/configfile.h>
#include <libplayerc++/playerc++.h>
#include <sensor/sensor_base/sensorreading.h>
#include <sensor/sensor_range/rangesensor.h>
#include <sensor/sensor_range/rangereading.h>


/* Include Qt files  */
#include <QtGui/QApplication>
#include <QtGui/QPainter>
#include <QtGui/QPixmap>
#include <QtGui/QWidget>
#include <QtGui/QWMatrix>
#include <QtGui/QImage>
#include <QtGui/QMouseEvent>
#include <QtCore/QTime>
#include <QtCore/QTextStream>
#include <QtCore/QThread>
#include <QtCore/QReadWriteLock>
#include <QtCore/QMutex>
#include <QtOpenGL/QGLWidget>


/* Include STL files */
#include <cstdio>
#include <string>
#include <vector>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <algorithm>
using namespace std;

#include "Map.h"

#include "camClass.h"

typedef struct vec3f
{
	float x;
	float y;
	float z;
} vec3f;

class RobotController;

class LaserViewer : public QGLWidget
{

 public:

  int lastX,lastY;
  float view_rot_x, view_rot_y;
  QMutex * glDisplayLock;
  
  LaserViewer();
  void setPts(vector<vec3f> pts);

 protected:
  vector<vec3f> pts;
  void initializeGL();
  void resizeGL( int, int );
  void paintGL();
  void mousePressEvent(QMouseEvent* e);
  void mouseMoveEvent(QMouseEvent *qme);
  void timerEvent(QTimerEvent* qte);
};


/* Typedefs to make life easy */
typedef OrientedPose      RobotPose;
typedef Map               ObstacleMap;
typedef Map               BoundaryMap;
typedef Map               PotentialMap;
typedef Map               DistanceMap;
typedef vector<RobotPose> WaypointList; 
typedef RobotPose         GridPoint;
typedef vector<GridPoint> GridPointList;
typedef vector<RobotPose> Contour;
typedef vector<Contour>   ContourList;
typedef vector<float>     LaserScanData;
typedef vector<bool>      VisitedWaypoints;

/* Possible robot states */
enum RobotState { START, RANDOM_WALK, ON_ROUTE, ARRIVED, FINISHED };

/* Walk times */
const int TOTAL_WALK_TIME     = 1000*60*15;
const int RANDOM_WALK_TIME    = 1000*60;
const int WAYPOINT_WALK_TIME  = TOTAL_WALK_TIME - RANDOM_WALK_TIME;

/* Robot Velocities */
const float ROT_VEL   = 30.0f*(M_PI)/180.0f;
const float TRANS_VEL = 0.08f;

/* Tilt Unit Parameters */
const int   NUM_TILT_ANGS   = 1;
const float TILT_ANG_OFF    = 32.0f*M_PI/180.0f;
const float TILT_ANG_STEP   = 10.0f*M_PI/180.0f;
const float TILT_ANG_MAX    = (int(NUM_TILT_ANGS/2)*TILT_ANG_STEP);
const float TILT_WHEEL_DIST = 0.12f;

/* Forward Declarations */
class SensorUpdateThread;
class GridSlamThread;
class WalkThread;
class CommandModeThread;

class RobotController : public QWidget
{
 public:
  void processCommandLine();

  CamClass *cams;
  void initializeCameras();

  bool useLaserViewer;
  LaserViewer* lv;
  QWidget    * rangeSensor2DViewer;

  QPainter* qp;            /* Qt brush class                                */
  float deltaX, deltaY;    /* Stretch between grid coords and screen coords */
  int  w,h;                /* Width and Height of current window            */
  bool don;                /* Use visualization                             */
  bool commandMode;        /* Give commands on command line                 */

  float        rr;       /* Radius of the robot         */
  //  RobotPose    rp;       /* Pose of the robot           */
  ObstacleMap  om;       /* Obstacle map seen by slam   */
  BoundaryMap  bm;       /* Boundary of obstacle map    */
  BoundaryMap  bm2;      /* Uncontourized boundary map  */
  PotentialMap pm;       /* Obstacle potential map      */
  //PotentialMap tm;       /* Target potential map        */

  unsigned int  cwp;     /* Index of current waypoint   */
  WaypointList wpl;      /* List of all waypoints       */

  RobotState       rs;   /* State of the robot          */
  PlayerClient    *rbt;  /* Player handle for the robot */
  Position2dProxy *pp;   /* Proxy to set robot speed    */
  PtzProxy        *ptzp; /* Proxy to control pan tilt   */
  LaserProxy      *lp;   /* Proxy to control laser unit */


  WaypointList cpth;     /* Current path of the robot   */
  VisitedWaypoints vwp;

  float mpres;           /* Resolution of the map       */

  RobotPose     lastSlamPoseWorld;         /* The last pose returned by slam in world coordinates */
  RobotPose     lastOdomPoseWorld;         /* The last pose returned by odom in world coordinates */
  RobotPose     lastSlamPoseLocal;        /* The last pose returned by slam in local coordinates */
  RobotPose     lastSlamOdomPoseWorld;    /* Odometry pose at time of last slam computation */
  
  RobotPose     currentRobotPose;

  RobotPose     currentWaypointPose;

  QReadWriteLock     *slck;        /* Read-Write Lock on sensors  */      
  QReadWriteLock     *gllck;       /* Read-Write Lock on range sensor transformed points  */     
  QReadWriteLock     *slamMapLock; /* Read-Write Lock on the slam map */
  
  SensorUpdateThread *sut;   /* Thread for polling sensors      */ 
  GridSlamThread     *gst;   /* Thread for running slam         */
  WalkThread         *wt;    /* Thread for running the walk     */
  CommandModeThread  *cmt;   /* Thread for running command line */

  int           lcnt;      /* Number of samples per laser scan */
  float         lres;      /* Angle between two samples        */
  LaserScanData lsd;       /* The laser scanning data          */


  GMapping::GridSlamProcessor *gsp;   /* The SLAM processor */
  GMapping::RangeSensor       *grs;   /* SLAM range sensor  */
  
  vector<vec3f> laserPts;
 public:

  RobotController();

  
  /* Planned Motion Functions */
  void doWalk        ();
  void doRandomWalk  (float rwt, bool bcaptureImages);
  void doWaypointWalk(float wwt);

  /* Misc. IO functions */
  int  readSLAMData              ();
  int  readOccupancyGridFromFile (string fileName);
  int  readRobotPoseFromFile     (string fileName);
  int  readLaserScanFromFile     (string fileName, vector<float> &lsd);

  /* High Level Planner functions */
  Map slamMap;

  void computeWaypointList                ();
  void computeBoundaryMap                 (float t, int md);
  void computeWaypointListFromContourList (const ContourList& cl);
  void computeContourListFromBoundaryMap  (ContourList& cl);
  void compressWaypointList               ();
  void moveWaypointsAwayFromObjects       (int a_Dist);
  bool waypointNearObstacle               (RobotPose &a_pose, int a_Distance);


  RobotPose getNextWaypoint(Map& tm);
  
  /* Low Level Planner functions   */
  void computePathToTarget      (RobotPose startPose, RobotPose tp, Map& tm);
  void computePotentialMap      ();
  void computeTargetDistanceMap (Map& m,RobotPose tp);  
  void computeShortestDistance  (DistanceMap& dm, GridPointList& gpl, bool upm=false);
  void computeObstacleAndPotentialMap();

  bool isClosePose  (RobotPose rp1, RobotPose rp2, bool cmpr = false, float minDist=2);	  
  
  /* GUI functions */
  void initializeGUI();
  void drawCurrentStatus();
  void paintEvent(QPaintEvent* qpe);
  void mousePressEvent(QMouseEvent* qme);
  void timerEvent(QTimerEvent* qte);
  void drawRobotPose(const RobotPose& rp);
  void drawMap      (const Map&       m);
  void drawGrid     (int   spacing);
  void drawInfo     ();
  

  void initializeRobot(string hn, int pn);

  /* Robot motion functions */
  void rotateInPlace  (float t, float a_Multiplier);
  void moveForward    (float d);
  bool moveToAtomic(RobotPose& start, RobotPose target, float& transAmount, float& rotAmount);
  bool moveTo      (RobotPose target);
  int moveToTarget(RobotPose target);


  /* Robot sensor functions */
  void triggerCameras();
  //  void updateMeasurements();
  void readCurrentMeasurements();
  bool readLaserScan(size_t sz, float fAngle, vector<vec3f>& pts);

  /* SLAM functions */
  void initializeGridSlam();
  void updateGridSlam();

  /* Laser Functions */
  vector<float> m_tiltAngles;
  bool Quantize( const vector<vec3f>& pts, vector<float>& OneScan );
  bool TransformLaserLocalToGlobal( const vector<vec3f>& ptsLocal, float Tx, float Ty, float theta, vector<vec3f>& ptsGlobal );
  bool TransformRadialTo3D( float fAngle, vector<vec3f>& pts );
  void updateMeasurements();
  void wheelsToLaserCoordinates(float curwXPos, float curwYPos, float curwTheta, float &curlXPos, float &curlYPos, float &curlTheta);
  void laserToWheelCoordinates(float curwXPos, float curwYPos, float curwTheta, float &curlXPos, float &curlYPos, float &curlTheta);


};

/* Thread class for starting sensor update loop */
class SensorUpdateThread : public QThread
{
  Q_OBJECT

 public:
  SensorUpdateThread(RobotController* rc) : rc(rc) {}
  
  virtual void run()    
    { 
      for (;;) 
	{ 
	  usleep(150000);// Sleep for 0.15 sec 
	  rc->updateMeasurements(); 
	} 
    }

 private:
  RobotController *rc;
};

/* Thread class for starting slam update loop */
class GridSlamThread : public QThread
{
  Q_OBJECT

 public:
  GridSlamThread(RobotController* rc) : rc(rc) {}
  
  virtual void run()    
    {
      for (;;) 
	{
	  rc->updateGridSlam();
	}
    }

 private:
  RobotController *rc;
};

class WalkThread : public QThread
{
  Q_OBJECT

 public:
  WalkThread(RobotController* rc) : rc(rc) {}
  
  virtual void run()    
    {
      rc->doWalk();
    }

 private:
  RobotController *rc;
};

class CommandModeThread : public QThread
{
  Q_OBJECT

 public:
  CommandModeThread(RobotController* rc) : rc(rc) {}
  
  virtual void run()    
    {
      for(;;) 
	{
	  rc->processCommandLine();
	}
    }

 private:
  RobotController *rc;
};

#endif
