#include "RobotController.h"

#include <map>
#include <queue>
#include <cmath>
#include <iostream>
using namespace std;


class PointDistancePair
{
public:
  int   x,y;
  float d;
  PointDistancePair(int x = 0, int y = 0, float d = 0) : x(x), y(y), d(d)                                  {}
  PointDistancePair(float x , float y , float d = 0) : x(int(x)), y(int(y)), d(d)                    {}
  bool operator() (const PointDistancePair &a, const PointDistancePair &b) const { return a.d>b.d; }
};



void RobotController::computePotentialMap()
{
  pm = om;
  
  GridPointList gpl;

  for (int x=0; x<om.xSize; x++)
    for (int y=0; y<om.ySize; y++)
      if (om(x,y) > 0)
	gpl.push_back(GridPoint(x,y));

  computeShortestDistance(pm, gpl);

  for (int x=0; x<om.xSize; x++)
    for (int y=0; y<om.ySize; y++)
      if (pm(x,y) < 1)
	pm(x,y)=1e6;
      else
	pm(x,y)=std::max(10.0f-pm(x,y),1.0f);
}

void RobotController::computeTargetDistanceMap(Map& m, RobotPose tp)
{
  
  GridPointList gpl;
  gpl.push_back(tp);

  computeShortestDistance(m,gpl,true);
}

void RobotController::computeShortestDistance(DistanceMap& sdm, GridPointList& gpl, bool upm)
{
//cout<<"computing shortest path"<<endl;
  float gds[3][3] = {{ sqrtf(2), 1, sqrtf(2)},
		     {    1    , 0, 1       },
		     { sqrtf(2), 1, sqrtf(2)}};

	                  
  priority_queue<PointDistancePair, vector<PointDistancePair>, PointDistancePair> pq;

  for (int x=0; x<sdm.xSize; x++)
    for (int y=0; y<sdm.ySize; y++)
      sdm(x,y) = 1e6;

  for (unsigned int i=0; i<gpl.size(); i++)
    {
      pq.push(PointDistancePair(gpl[i].x,gpl[i].y,0));
      sdm(gpl[i].x, gpl[i].y) = 0;
    }
	
  while (!pq.empty())
    {
    //cout<<"popping nearest"<<endl;

      PointDistancePair pdp = pq.top(); pq.pop();
      int   curx = pdp.x;
      int   cury = pdp.y;
      float ndist = 0;
    //cout<<" considering "<<curx<<" "<<cury<<pdp.d<<endl;	 
      for (int i=-1; i<=1; i++)
	for (int j=-1; j<=1; j++)
	  {
	    if (curx+i < 0)                 continue;
	    if (curx+i > sdm.xSize-1)       continue;
	    if (cury+j < 0)                 continue;
	    if (cury+j > sdm.ySize-1)       continue;

	    if (om(curx+i, cury+j) > 0) continue;

	    if (upm)
	      ndist = pdp.d + gds[i+1][j+1]*(pm(curx,cury)+pm(curx+i,cury+j))/2;
	    else
	      ndist = pdp.d + gds[i+1][j+1];
	    
	    if (sdm(curx+i, cury+j) > ndist)
	      {
		sdm(curx+i,cury+j)  = ndist;
		pq.push(PointDistancePair(curx+i, cury+j, sdm(curx+i,cury+j)));
	      }
	  }
    }
}

void RobotController::computePathToTarget(RobotPose startPose, RobotPose tp, Map& tm)
{
  cpth.clear();
  RobotPose crp = startPose;
  cerr<<"computing path to target"<<endl;
  
  while (!isClosePose(crp,tp,false,1))
    {
      int x = int(crp.x);
      int y = int(crp.y);
      float mnpot = tm(x,y);
      //cout<<" looking at "<<x<<" "<<y<<endl;
      //int k;
      //cin>>k;
      for (int i=-1; i<=1; i++)
	for (int j=-1; j<=1; j++)
	  {
	    if (x+i < 0)                 continue;
	    if (x+i > tm.xSize-1)        continue;
	    if (y+j < 0)                 continue;
	    if (y+j > tm.ySize-1)        continue;
	    if (i==0 && j==0)            continue;
	    if (tm(x+i,y+j)<mnpot)
	      {
		mnpot = tm(x+i,y+j);
		crp = RobotPose(x+i,y+j);
		cpth.push_back(crp);
	      }
	    
	  }
      
    }
}


bool RobotController::moveToAtomic(RobotPose& startPose, RobotPose tp, float& transAmount, float& rotAmount)
{
  cerr<<"  In Move to atomic"<<endl;
//  cerr<<"   Current pose "<<startPose.x<<" "<<startPose.y<<" "<<startPose.t<<endl;
//  cerr<<"   Target  pose "<<tp.x<<" "<<tp.y<<" "<<tp.t<<endl;
  
  RobotPose rp = startPose;

  /* Compute relative distance and angle in world lengths */
  float d = sqrtf(pow(tp.y-rp.y,2)+pow(tp.x-rp.x,2))*mpres;
  float t = atan2(tp.y-rp.y,tp.x-rp.x) - rp.t;
  t = fmod((t + M_PI),(2*M_PI)) - M_PI;

  cerr<<"     Calculated d="<<d<<" t="<<t<<endl;
  usleep(50000);
  /* Execute motion */
  rotateInPlace(-t,1.0f);
  cerr<<"     rotated"<<endl;
  usleep(50000); //needed so that robot does not get stuck in rotation
  moveForward(d);
  cerr<<"     translated"<<endl;
  /* Make target pose the current pose */
  startPose = tp;
  startPose.t = t+rp.t;
  //t = (tp.t-t);
  //rotateInPlace(-t.0f);
  transAmount = d;
  rotAmount =t;
  //currentRobotPose = rp;
  return true;
}

int RobotController::moveToTarget(RobotPose targetPose)
{
    RobotPose localCurrentRobotPose = currentRobotPose;

//    cerr<<"Moving to "<<targetPose.x<<" "<<targetPose.y<<" "<<targetPose.t<<endl;
//    cerr<<"From "<<localCurrentRobotPose.x<<" "<<localCurrentRobotPose.y<<" "<<localCurrentRobotPose.t<<endl;
    
    Map targetDistanceMap;
    targetDistanceMap = om;

    computeTargetDistanceMap(targetDistanceMap, targetPose);

    if (om(localCurrentRobotPose.x,localCurrentRobotPose.y)==1)
    {
        cout<<" Help! I think I am an obstacle "<<endl;
        return -1;
    }
    
    if (targetDistanceMap(localCurrentRobotPose.x, localCurrentRobotPose.y)>1e5)
    {
        cout<<" No path to that waypoint, skipping"<<endl;
        return -2;
    }        

    computePathToTarget(localCurrentRobotPose, targetPose, targetDistanceMap);
/*    cerr<<" Computed Path :";
        
    for(unsigned int i=0; i<cpth.size(); i++)
        cerr<<"("<<cpth[i].x<<","<<cpth[i].y<<")->";

    cerr<<endl;
*/
    int nextPathPoint = 0;
    
    float totalTransAmount=0, totalRotAmount=0;

    while(!isClosePose(targetPose, localCurrentRobotPose, false, 2))
    {
        // Define how many path points to skip
        nextPathPoint = nextPathPoint+std::min(8, int(cpth.size()-1)-nextPathPoint); // TODO: Change the skip pixels from 6 to 8

//        cerr<<" moving to next pose toward target "<<cpth[nextPathPoint].x<<" "<<cpth[nextPathPoint].y<<endl;

        float transAmount=0, rotAmount =0;
        moveToAtomic(localCurrentRobotPose, cpth[nextPathPoint], transAmount, rotAmount);

        usleep(50000);
        totalTransAmount += transAmount;
        totalRotAmount += rotAmount;

        if (totalTransAmount > 0.35 || totalRotAmount> 30.0f*M_PI/180.0f)
        { 
//          cout<<"sleeping, then recomputing path"<<endl;
          sleep(2);
          computeObstacleAndPotentialMap();
//        cerr<<"recomputing waypoints"<<endl;
          //computeWaypointList();
            //for (unsigned int i=0; i<wpl.size(); i++)
              //  cerr<<" Waypoint #"<<i<<" = "<<wpl[i].x<<" "<<wpl[i].y<<" "<<wpl[i].t<<endl;
//          cerr<<"updating current robot pose"<<endl;
 	      slamMapLock->lockForRead();
	      slck->lockForRead();

	      currentRobotPose.x = lastSlamPoseLocal.x + int((lastOdomPoseWorld.x-lastSlamOdomPoseWorld.x)/mpres);
	      currentRobotPose.y = lastSlamPoseLocal.y + int((lastOdomPoseWorld.y-lastSlamOdomPoseWorld.y)/mpres);
	      currentRobotPose.t = fmod(lastSlamPoseLocal.t + lastOdomPoseWorld.t-lastSlamOdomPoseWorld.t , float(2*M_PI) );
//	      cerr<<"currentRobotPose updated to "<<currentRobotPose.x<<" "<<currentRobotPose.y<<" "<<currentRobotPose.t<<endl;
	      slck->unlock();
	      slamMapLock->unlock();
          localCurrentRobotPose = currentRobotPose;

          targetDistanceMap = om;
          computeTargetDistanceMap(targetDistanceMap, targetPose);
            
          if (om(localCurrentRobotPose.x,localCurrentRobotPose.y)==1)
          {
             cout<<" Help! I think I am an obstacle "<<endl;
             return -1;
          }
    
          if (targetDistanceMap(localCurrentRobotPose.x, localCurrentRobotPose.y)>1e5)
          {
             cout<<" No path to that waypoint, skipping"<<endl;
             return -2;
          }        

          computePathToTarget(localCurrentRobotPose, targetPose, targetDistanceMap);
//          cerr<<" Computed Path :";
        
          for(unsigned int i=0; i<cpth.size(); i++)
             cerr<<"("<<cpth[i].x<<","<<cpth[i].y<<")->";

          cerr<<endl;

          nextPathPoint = 0;

          totalTransAmount = 0;
          totalRotAmount = 0;

        }
        cerr<<" Move complete"<<endl;
 
        //localCurrentRobotPose = cpth[nextPathPoint];       NOT NEEDED moveToAtomic is changing localCurrentRobotPose
    }

   
    float t = (targetPose.t-localCurrentRobotPose.t);
//    t = fmod((t + M_PI),(2*M_PI)) - M_PI;
    
//    rotateInPlace(-t);

//    cerr<<"rotating to look at the waypoint\n"<<endl;
    t = fmod((t + M_PI),(2.0f*M_PI)) - M_PI; // We want to capture multiple images, so the initial position is t-40 degrees
    rotateInPlace(-t+80.0f*M_PI/180.0f,2.0f);
//    cerr<<"rotating to look at the waypoint-80 degrees\n"<<endl;         
//    cerr << "Reached target" <<endl;
//    cerr<<"Stopping tour at "<<localCurrentRobotPose.x<<" "<<localCurrentRobotPose.y<<" "<<localCurrentRobotPose.t<<endl;
   
    return 1;
}
    


