#include "RobotController.h"

/* Initialize gmapping   */
void RobotController::initializeGridSlam()
{
  cout<<"Initializing GridSlam"<<endl;
  
  /* Map parameters */
  double xmin=-10.;
  double ymin=-10.;
  double xmax=10.;
  double ymax=10.;
  double delta=mpres; 

  /* Scan matching parameters */
  double sigma=0.05;
  double maxrange=4;
  double maxUrange=4;
  double lstep=.05;
  double astep=.05;

  int lskip=0;  
  int kernelSize=1;
  int iterations=5;

  double lsigma=.075;
  double ogain=3;

  /* Motion model parameters */
  double srr=0.01, srt=0.01, str=0.01, stt=0.01;

  /* Number of particles (map hypotheses) to store */
  int particles=20;
  
  
  /* GFS parameters */
  double angularUpdate=35.0*3.14/180.0;//25*3.1416/180.0;
  double linearUpdate=0.25; // 0.35
  double resampleThreshold=0.5;
  bool   generateMap=true;
  
  gsp = new GMapping::GridSlamProcessor;
  string sensorName="FLASER";
  
  while (lsd.size()==0)
    {
      cout<<"   Waiting for laser to initalize"<<endl;
      usleep(1000000);
    }
    
  grs = new GMapping::RangeSensor("FLASER", lcnt, lres, GMapping::OrientedPoint(0,0,0), 0,maxUrange);
  grs->updateBeamsLookup();
  GMapping::SensorMap m_sensorMap;
  m_sensorMap.insert(make_pair(sensorName, grs));
  
  gsp->setSensorMap(m_sensorMap);
	
  /* Set the command line parameters */
  gsp->setMatchingParameters(maxUrange, maxrange, sigma, kernelSize, lstep, astep, iterations, lsigma, ogain, lskip);
  gsp->setMotionModelParameters(srr, srt, str, stt);
  gsp->setUpdateDistances(linearUpdate, angularUpdate, resampleThreshold);
  gsp->setgenerateMap(generateMap);

  float curlXPos=0., curlYPos=0., curlTheta=0.;
  wheelsToLaserCoordinates(lastOdomPoseWorld.x, lastOdomPoseWorld.y,lastOdomPoseWorld.t, curlXPos, curlYPos, curlTheta);
  GMapping::OrientedPoint initialPose(curlXPos, curlYPos, curlTheta); 
  
  /* GridSlamProcessor initialization */
  gsp->init(particles, xmin, ymin, xmax, ymax, delta, initialPose);
  cout<<"   Initializing GridSlamProcessor with initialPose="<<initialPose.x<<" "<<initialPose.y<<" "<<initialPose.theta<<endl;
  
  /* Set last slam pose in world coordinates*/
  lastSlamPoseWorld.x = lastOdomPoseWorld.x;
  lastSlamPoseWorld.y = lastOdomPoseWorld.y;
  lastSlamPoseWorld.t = lastOdomPoseWorld.t;
  

}


/* One iteration of slam */
void RobotController::updateGridSlam()
{

  GMapping::RangeReading reading(grs, time(NULL));
  reading.resize(grs->beams().size());
 
  slck->lockForRead();
  
  for (int i=0; i< lcnt; i++)
    {
      reading[i]=lsd[lcnt-i-1];
    }
  
  float curlXPos=0., curlYPos=0., curlTheta=0.;
  wheelsToLaserCoordinates(lastOdomPoseWorld.x, lastOdomPoseWorld.y,lastOdomPoseWorld.t, curlXPos, curlYPos, curlTheta);
  /* Give the current odometry position to SLAM */
  reading.setPose(GMapping::OrientedPoint(curlXPos,curlYPos,curlTheta));

  slck->unlock();
  
  
  bool processed=gsp->processScan(reading);
//    cout << "SLAM ITERATION" << endl; 
  /* This returns true when the algorithm effectively processes (the traveled path since the last processing is over a given threshold) */
  if (processed){
    slck->lockForRead();
    lastSlamOdomPoseWorld = lastOdomPoseWorld;
    slck->unlock();
    
    /* Get index of the best map */
    unsigned int best_idx=gsp->getBestParticleIndex();
    
    /* Copy to a normal map */    
    //     GMapping::Map<double, GMapping::DoubleArray2D, false>* mymap = gsp->getParticles()[best_idx].map.toDoubleMap();
    GMapping::ScanMatcherMap mymap = gsp->getParticles()[best_idx].map;
    GMapping::Point pmin=mymap.map2world(GMapping::IntPoint(0,0));
    GMapping::Point pmax=mymap.map2world(mymap.getMapSizeX()-1,mymap.getMapSizeY()-1);
    GMapping::Point center=(pmax+pmin)*0.5;
    GMapping::Map<double, GMapping::DoubleArray2D, false>*  plainMap=new GMapping::Map<double, GMapping::DoubleArray2D, false>(center, (pmax-pmin).x, (pmax-pmin).y, mymap.getDelta());
    for(int x=0; x<plainMap->getMapSizeX(); x++)
      for(int y=0; y<plainMap->getMapSizeY(); y++){
    	GMapping::IntPoint p(x,y);
    	plainMap->cell(p)=mymap.cell(p);
      }
    GMapping::OrientedPoint robot_pose=(gsp->getParticles()[best_idx]).pose;
    
    
//    cout<<"Map size is "<<mymap.getMapSizeX()<<" "<<mymap.getMapSizeY()<<endl;
//    cout<<"Other Map size is "<<plainMap->getMapSizeX()<<" "<<plainMap->getMapSizeY()<<endl;
    

    slamMapLock->lockForWrite();
//    cout<<"slam map locked"<<endl;
    slamMap.resize(plainMap->getMapSizeX(),plainMap->getMapSizeY());
    for(int x=0; x<plainMap->getMapSizeX(); x++)
      for (int y=0; y<plainMap->getMapSizeY(); y++)
    	slamMap(x,y) = plainMap->cell(x,y);
    
    laserToWheelCoordinates(robot_pose.x, robot_pose.y, robot_pose.theta, lastSlamPoseWorld.x, lastSlamPoseWorld.y, lastSlamPoseWorld.t);
    
    GMapping::IntPoint int_pose(plainMap->world2map(GMapping::Point(robot_pose.x,robot_pose.y)));
    
    
    lastSlamPoseLocal.x = int_pose.x;
    lastSlamPoseLocal.y = int_pose.y;
    lastSlamPoseLocal.t = robot_pose.theta;
    
    
    
    currentRobotPose  = lastSlamPoseLocal;
//    cout<<"last local theta="<<lastSlamPoseLocal.t<<endl;
    slamMapLock->unlock();
//    cout<<"slam map unlocked"<<endl;
    cout<<"GridSlam updated the map"<<endl;
  } 
}
